;; Major mode for editing PoVray scene Files
;; 
;; NOTE: To achieve any sort of reasonable performance, YOU MUST
;;   byte-compile this package.  In emacs, type M-x byte-compile
;;   and then enter the name of this file.
;;
;; Only tested on Emacs 19 and XEmacs 19.  Let me know about
;;  any compatibility problems.
;; 
;; Modifief by: Francois Dispot <wozzeck@club-internet.fr>
;;  5/06/00:
;;    Added speedbar support
;; Modifief by: Jerome Berger <bergerj@iname.com>
;;  4/19/00:
;;    Added support for MegaPov 0.4
;;  5/12/99:
;;    Added support for Pov 3.1 fontification
;;    Note that I didn't add support for completion
;; Modified by: Peter Boettcher <pwb@andrew.cmu.edu>
;;  5/8/97:
;;    Added font-lock support for Emacs/XEmacs 19
;;    Indent under `#declare Object=' lines
;;    Corrected comment syntax
;;    Got rid of more remnants from postscript mode
;;    General cleanup
;;    Arbitrarily chose version 1.2
;; 5/8/97:  Version 1.21
;;    fontify-insanely was ignored.  fixed.
;;
;; 9/24/97: Version 1.3
;;    Added indentation for Pov 3 syntax (#if #else, etc)
;;    Preliminary context-sensitive keyword completion
;;
;; Original Author:	Kevin O. Grover <grover@isri.unlv.edu>
;; 	  Cre Date:	04 March 1994
;; This file derived from postscript mode by Chris Maio
;;
;;
;;  Please send bug reports/comments/suggestions to me at
;;        pwb@andrew.cmu.edu
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; The following two statements, placed in your .emacs file or site-init.el,
;; will cause this file to be autoloaded, and pov-mode invoked, when
;; visiting .pov files:
;;
;;	(autoload 'pov-mode "pov-mode.el" "PoVray scene file mode" t)
;;	(setq auto-mode-alist
;;	      (cons '("\\.pov$" . pov-mode) auto-mode-alist)
;;            (cons '("\\.inc$" . pov-mode) auto-mode-alist))
;; 
;; Use this to turn on font-lock:
;;	(add-hook 'pov-mode-hook 'turn-on-font-lock)
;;;;;;;;;;;;;;;;;;;; 
;; These variables can be set in your .emacs (defaults given):
;; (setq pov-indent-level '2)
;; (setq pov-autoindent-endblocks t)   Automatically reindents else/end/break
;; (setq pov-indent-under-declare '2)  Try it!  Tell me if you like it...
;; (setq pov-fontify-insanely t)       When it's non-nil, we fontify *every*
;;                                     Povray keyword.  Careful!
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  TODO list:
;;     Link to povray documentation
;;     Vector operations (add <0, .5, 1> to every vector in region)
;;     Clean up completion code
;;     Do something smarter than the default for compilation mode
;;     TAGS, to jump to #declared objects elsewhere in the code
;;
;;   Feel free to do any of this and send me code :)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'pov-mode)
(require 'easymenu)
(require 'imenu)
(require 'speedbar)

(defvar pov-xemacsp (string-match "XEmacs" emacs-version)
  "Non-nil if we are running in the XEmacs environment.")

(defvar pov-menu
  (list "POV"
	["Speedbar"          speedbar-frame-mode t]
	)
  "Menu for JDE.")

(defconst pov-mode-version '1.3.1)

(defvar pov-indent-level 2
  "*Indentation to be used inside of PoVray blocks or arrays")

(defvar pov-autoindent-endblocks t
  "*When non-nil, automatically reindents when you type break, end, or else")

(defvar pov-indent-under-declare 2
  "*Indentation under a `#declare Object=' line")

(defvar pov-fontify-insanely t
  "*Non-nil means colorize every povray keyword.  This may take a while on lare files.  Maybe disable this on slow systems.")

(defconst pov-tab-width 8
  "*Tab stop width for PoV mode")

(defun pov-make-tabs (stop)
  (and (< stop 132) (cons stop (pov-make-tabs (+ stop pov-tab-width)))))

(defconst pov-tab-stop-list (pov-make-tabs pov-tab-width)
  "Tab stop list for PoV mode")

(defconst pov-command "povray -"
  "Command used to invoke the povray program.")

(defvar pov-mode-map nil
  "Keymap used in PoV mode buffers")

(defvar pov-mode-syntax-table nil
  "PoV mode syntax table")

(defconst pov-comment-start-regexp "//\\|/\\*"
  "Dual comment value for `comment-start-regexp'.")

(defvar pov-comment-syntax-string ". 124b"
  "PoV hack to handle Emacs/XEmacs foo")

(defvar pov-begin-re "\\<#\\(if\\(n?def\\)?\\|case\\|macro\\|range\\|switch\\|while\\)\\>")
(defvar pov-end-re  "\\<#break\\|#end\\>")

(defvar pov-else-re "\\<#else\\>")

(defvar pov-begin-end-re (concat
			  pov-begin-re
			  "\\|"
			  pov-end-re
			  "\\|"
			  pov-else-re))

(defun pov-setup-syntax-table nil
  (if (or (string-match "Lucid" emacs-version)
	  (string-match "XEmacs" emacs-version))
      (setq pov-comment-syntax-string ". 1456"))
  (if pov-mode-syntax-table
      ()
    (setq pov-mode-syntax-table (make-syntax-table))
    (modify-syntax-entry ?_ "w" pov-mode-syntax-table)
    (modify-syntax-entry ?# "w" pov-mode-syntax-table)
    (modify-syntax-entry ?/ pov-comment-syntax-string pov-mode-syntax-table)
    (modify-syntax-entry ?* ". 23" pov-mode-syntax-table)
    (modify-syntax-entry ?\n "> b" pov-mode-syntax-table)
    (modify-syntax-entry ?< "." pov-mode-syntax-table)
    (modify-syntax-entry ?> "." pov-mode-syntax-table)
    (modify-syntax-entry ?= "." pov-mode-syntax-table)
    (modify-syntax-entry ?+ "." pov-mode-syntax-table)
    (modify-syntax-entry ?- "." pov-mode-syntax-table)
    (modify-syntax-entry ?. "w" pov-mode-syntax-table)
    (set-syntax-table pov-mode-syntax-table)))

; Huge regexp for pov-keyword hilighting.  Organized alphabetically.  Try
; to read at your own risk!
(defconst pov-a-keywords-matcher
  '("\\<a\\(a_\\(level\\|threshold\\)\\|bs\\|bsorption\\|ccuracy\\|cosh?\\|d\\(aptive\\|c_bailout\\)\\|gate\\(_turb\\)?\\|l\\(ign_\\(center\\|left\\|right\\)\\|l\\(_intersections\\)?\\|pha\\|ways_sample\\)\\|mbient\\(_light\\)?\\|ngle\\|p\\(erture\\|pend\\)\\|r\\(c_angle\\|ea_light\\|ray\\)\\|sc\\|sinh?\\|ssumed_gamma\\|t\\(an[2h]?\\|mospher\\(e\\|ic_attenuation\\)\\|tenuating\\)\\|utostop\\|verage\\)\\>" . font-lock-keyword-face))
(defconst pov-b-keywords-matcher
  '("\\<b\\(ackground\\|ands\\|l\\(ack_hole\\|inn\\|ob_pigment\\|otches\\|ue\\|ur_samples\\)\\|o\\(unded_by\\|x\\(_mapping\\|ed\\)\\|zo\\)\\|r\\(eak\\|ick\\(_size\\)?\\|ightness\\|illiance\\)\\|_spline_sphere_sweep\\|ump\\(s\\|y[123]?\\|_map\\|_size\\)\\)\\>" . font-lock-keyword-face))
(defconst pov-c-keywords-matcher
  '("\\<c\\(a\\(bs2\\|cosh?\\|mera\\|rtoon\\|se\\|sinh?\\|tanh?\\|tmull_rom_spline_sphere_sweep\\|ustics\\)\\|co\\(nj\\|sh?\\)\\|div\\|e\\(il\\|lls\\|xp\\)\\|h\\(ecker\\|r\\)\\|imag\\|ircular\\|l\\(ipped_by\\|o\\(ck\\(_delta\\)?\\|g\\|sed?\\th2?\\)\\)\\|mult\\|o\\(lou?r\\(_map\\)?\\|mpo\\(nent\\|site\\)\\|n\\(cat\\|fidence\\|ic_sweep\\|serve_energy\\|stant\\|tained_by\\|trol[01]\\)\\|ords\\|sh?\\|unt\\)\\|pow\\|ra\\(ckle\\|nd\\)\\|real\\|s\\(inh?\\|qrt?\\)\\|tanh?\\|u\\(b\\(ic_wave\\)?\\|rrent_chrono\\)\\|ylindrical\\(_mapping\\)?\\)\\>" . font-lock-keyword-face))
(defconst pov-d-keywords-matcher
  '("\\<d\\(ate\\|e\\(grees\\|n\\(\\(sity\\(_\\(file\\|function\\|map\\)\\)?\\)\\|ts\\)\\|pth\\|stroy\\)\\|f3\\|i\\(ff\\(erence\\|use\\)\\|m\\(ension\\(s\\|_size\\)\\)?\\|rection\\|s\\(p\\(_nelems\\|ersion\\)\\|tance\\(_maximum\\)?\\)\\|v\\)\\|ouble_illuminate\\|ust\\(_type\\)?\\)\\>" . font-lock-keyword-face))
(defconst pov-e-keywords-matcher
  '("\\<e\\(ccentricity\\|m\\(pty\\|i\\(tting\\|ssion\\)\\)\\|rror_bound\\|val\\(_\\(\\(3d_\\)?spline\\pattern\\|pigment\\)\\)?\\|x\\(p\\(and_thresholds\\|onent\\)?\\|tinction\\)\\)\\>" . font-lock-keyword-face))
(defconst pov-f-keywords-matcher
  '("\\<f\\(a\\(ce\\(_indices\\|ts\\)\\|de_\\(colou?r\\|distance\\|power\\)\\|l\\(loff\\(_angle\\)?\\|se\\)\\)\\|i\\(l\\(e_exists\\|ter\\)\\|n\\(al_\\(frame\\|clock\\)\\|ish\\)\\|sheye\\)\\|l\\(atness\\|ip\\|o\\(at\\|or\\)\\)\\|o\\(cal_\\(blur\\|point\\)\\|g\\(_alt\\|_offset\\|_type\\)?\\|rm\\)\\|r\\(actal_\\(exterior\\|interior\\)_type\\|ame_number\\|equency\\)\\|unc\\(3d\\|_id\\|tion\\|_xy\\)?\\)\\>" . font-lock-keyword-face))
(defconst pov-g-keywords-matcher
  '("\\<g\\(ather\\(_radius_step\\|_num_steps\\)?\\|if\\|lo\\(bal\\(_settings\\)?\\|wing\\)\\|r\\(adient\\|anite\\|ay_threshold\\|een\\|oups\\)\\)\\>" . font-lock-keyword-face))
(defconst pov-h-keywords-matcher
  '("\\<h\\(_angle\\|alo\\|exagon\\|f_\\(gray_16\\|height_at\\)\\|ierarchy\\|ollow\\|ypercomplex\\)\\>" . font-lock-keyword-face))
(defconst pov-i-keywords-matcher
  '("\\<i\\(dentifier\\|ff\\|gnore_photons\\|m\\(age_\\(height\\|map\\|pattern\\|width\\)\\|plicit\\)\\|n\\(cidence\\|i\\(_option\\|tial_\\(frame\\|clock\\)\\)\\|side_vector\\|t\\(er\\(polate\\|section\\|vals\\|ior\\(_texture\\)?\\)\\|_f\\)?\\|verse\\)\\|or\\|rid\\(_wavelength\\)?\\)\\>" . font-lock-keyword-face))
(defconst pov-j-keywords-matcher
  '("\\<j\\(itter\\|ulia[34]?\\)\\>" . font-lock-keyword-face))
(defconst pov-k-keywords-matcher nil)
(defconst pov-l-keywords-matcher
  '("\\<l\\(abel\\|ambda\\|eopard\\|i\\(b\\(_name\\|rary\\)\\|ght_\\(group\\|source\\)\\|near\\(_spline\\|\\(_sphere\\)?_sweep\\)?\\)\\|n\\|o\\(ad_file\\|cation\\|g\\|o\\(ad_file\\|k\\(s_like\\|_at\\)\\)\\|w_error_factor\\)\\)\\>" . font-lock-keyword-face))
(defconst pov-m-keywords-matcher
  '("\\<m\\(a\\(gnet[12][jm]\\|jor_radius\\|ndel[34]?\\|p_type\\|rble\\|terial\\(_map\\)?\\|trix\\|x\\(_density\\|_extent\\|_gather_count\\|_grad\\(ient\\)?\\|_intersections\\|_iteration\\|_sample\\|_trace\\(_level\\)?\\|_value\\)?\\)\\|e\\(dia\\(_attenuation\\|_interaction\\)?\\|rge\\|tallic\\|thod\\|tric\\)\\|in\\(imum_reuse\\|_extent\\|_gather\\(_count\\|_radius\\)\\)?\\|od\\(ify\\)?\\|ortar\\|otion_blur2?\\)\\>" . font-lock-keyword-face))
(defconst pov-n-keywords-matcher
  '("\\<n\\(atural\\|earest_count\\|o\\(_bump_scale\\|_eval\\|_image\\|ise3d\\|_reflection\\|rmal\\(_map\\|_vectors\\|_indices\\)?\\|_shadow\\)?\\|umber_of_waves\\)\\>" . font-lock-keyword-face))
(defconst pov-o-keywords-matcher
  '("\\<o\\(bject\\|ctaves\\|ff\\(set\\)?\\|mega\\|mnimax\\|n\\(ce\\|ion\\)?\\|pen\\|r\\(ient\\(ation\\)?\\|thographic\\)\\)\\>" . font-lock-keyword-face))
(defconst pov-p-keywords-matcher
  '("\\<p\\(a\\(noramic\\|rallel\\|ss_through\\|ttern[123]?\\)\\|erspective\\|gm\\|h\\(ase\\|ong\\(_size\\)?\\|otons\\)\\|i\\(gment\\(_map\\)?\\)?\\|lanar\\(_mapping\\)?\\|ng\\|o\\(int_at\\|larical\\|ly_wave\\|st_process\\|[tw]\\)\\|pm\\|re\\(cision\\|compute\\|trace_\\(start\\|end\\)\\)\\|rojected_through\\|roximity\\|wr\\)\\>" . font-lock-keyword-face))
(defconst pov-q-keywords-matcher
  '("\\<qu\\(adratic_\\(map\\|spline\\)?\\|aternion\\|ick_colou?r\\|ilted\\)\\>" . font-lock-keyword-face))
(defconst pov-r-keywords-matcher
  '("\\<r\\(a\\(di\\(al\\|ans\\|osity\\|us\\)\\|inbow\\|mp_wave\\|nd\\|nge\\(_divider\\)?\\|tio\\(nal\\)?\\)\\|e\\(ciprocal\\|cursion_limit\\|d\\|flect\\(ion\\(_blur\\(_max\\(_adc\\)?\\)?\\|_falloff\\|_m\\(ax\\|in\\)\\|_samples\\|_type\\)?\\|_metallic\\)\\(_exponent\\)?\\|fraction\\|peat\\|set_children\\)\\|gbf?t?\\|ight\\|ipples\\|otate\\|oughness\\)?\\>" . font-lock-keyword-face))
(defconst pov-s-keywords-matcher
  '("\\<s\\(amples?\\|ave_file\\(_bailout\\|s\\)\\|ave_file\\|ca\\(le\\|llop_wave\\|ttering\\)\\|eed\\|eparation\\|hadowless\\|heet\\|i\\(des\\|gn\\|n\\(e_wave\\|h\\)?\\|ze\\)\\|ky\\(_sphere\\)?\\|lice\\|lope\\(_map\\)?\\|mooth\\|o\\(ft_glow\\|lid\\)\\|p\\(acing\\|ecular\\|her\\(e_sweep_depth_tolerance\\|ical\\)\\(_camear\\|_mapping\\)?\\|iral[12]?\\|li\\(ne\\|t_union\\)\\|otlight\\|otted\\)\\|q\\(rt?\\|uare\\)\\|t\\(art_\\(chrono\\|end_direction\\)\\|eps\\|r\\(cmp\\|ength\\|ing\\(_encoding\\)?\\|len\\|lwr\\|upr\\)?\\|urm\\)\\|ubstr\\|witch\\|ys\\)?\\>" . font-lock-keyword-face))
(defconst pov-t-keywords-matcher
  '("\\<t\\(anh?\\|rget\\|est_camera_[1234]\\|exture\\(_map\\|_list\\)?\\|ga\\|hickness\\|hreshold\\|i\\(ck_count\\|ghtness\\)\\|ile[2s]\\|or\\(oidal\\|spiral\\)\\|r\\(ac[ek]\\|ans\\(form\\|late\\|mit\\)\\|iang\\(le_wave\\|ular\\)\\|immed_by\\|ue\\)\\|tf\\|urb\\(ulence\\|_depth\\)\\|ype\\)?\\>" . font-lock-keyword-face))
(defconst pov-u-keywords-matcher
  '("\\<u\\(ltra_wide_angle\\|n\\(ion\\|official_version\\)\\|p\\|se_\\(alpha\\|colou?r\\|index\\)\\|_steps\\|v_\\(indices\\|mapping\\|vectors\\)\\)?\\>" . font-lock-keyword-face))
(defconst pov-v-keywords-matcher
  '("\\<v\\(a\\(l\\|riance\\|xis_rotate\\)\\|cross\\|dot\\|ector\\|ertex_vectors\\|length\\|normalize\\|olume_\\(object\\|rendered\\)\\|ol_with_light\\|rotate\\|_steps\\|transform\\|turbulence\\)?\\>" . font-lock-keyword-face))
(defconst pov-w-keywords-matcher
  '("\\<w\\(a\\(rp\\|ter_level\\|ves\\)\\|idth\\|ood\\|rinkles\\)\\>" . font-lock-keyword-face))
(defconst pov-x-keywords-matcher
  '("\\<x\\>" . font-lock-keyword-face))
(defconst pov-y-keywords-matcher
  '("\\<y\\(es\\)?\\>" . font-lock-keyword-face))
(defconst pov-z-keywords-matcher
  '("\\<z\\>" . font-lock-keyword-face))

(defconst pov-all-keyword-matcher
  (list pov-a-keywords-matcher pov-b-keywords-matcher 
	pov-c-keywords-matcher pov-d-keywords-matcher 
	pov-e-keywords-matcher pov-f-keywords-matcher 
	pov-g-keywords-matcher pov-h-keywords-matcher 
	pov-i-keywords-matcher pov-j-keywords-matcher 
	;;pov-k-keywords-matcher 
	pov-l-keywords-matcher 
	pov-m-keywords-matcher pov-n-keywords-matcher 
	pov-o-keywords-matcher pov-p-keywords-matcher 
	pov-q-keywords-matcher pov-r-keywords-matcher 
	pov-s-keywords-matcher pov-t-keywords-matcher 
	pov-u-keywords-matcher pov-v-keywords-matcher 
	pov-w-keywords-matcher pov-x-keywords-matcher 
	pov-y-keywords-matcher pov-z-keywords-matcher))

(defvar pov-font-lock-keywords
  '(
    ("^[ \t]*\\(#declare\\)\\>[ \t]*\\(\\sw+\\)"  2 font-lock-variable-name-face nil t)
    ("\\<#\\(break\\|case\\|create\\|de\\(bug\\|clare\\|f\\(ault\\|close\\|ined\\)\\)\\|e\\(lse\\|nd\\|rror\\)\\|if\\(n?def\\)?\\|in\\(clude\\|it_\\(3d_\\)?spline\\)\\|local\\|macro\\|r\\(ange\\|ead\\|ender\\)\\|s\\(tatistics\\|witch\\)\\|undef\\|version\\|w\\(arning\\|hile\\|rite\\)\\)\\>" . font-lock-function-name-face)
    ("\\<\\(b\\(ezier_\\(patch\\|spline\\)\\|icubic_patch\\|lob\\|ox\\)\\|c\\(on\\(e\\|ic\\)\\|ub\\(e\\|ic\\(_\\(spline\\|wave\\)\\)?\\)\\|ylinder\\)\\|disc\\|height_field\\|iso\\(blob\\|surface\\)\\|julia_fractal\\|l\\(athe\\|inear_spline\\)\\|mesh2?\\|p\\(arametric\\|lane\\|oly\\(gon\\)?\\|rism\\)\\|qua\\(dratic\\(_spline\\)?\\|dric\\|rtic\\)\\|s\\(mooth_triangle\\|or\\|phere\\(_sweep\\)?\\|uperellipsoid\\)\\|t\\(ext\\|orus\\|riangle\\)\\|Rounded\\(Box\\|Cylinder\\)\\)\\>" . font-lock-type-face))
  "Expressions to highlight in PoV mode.")

(defun pov-setup-font-lock nil
  "Find out if we're fontifying insanely"
  (if pov-fontify-insanely
      (setq pov-font-lock-keywords (append pov-all-keyword-matcher pov-font-lock-keywords))))

(defun pov-mode nil
  "Major mode for editing PoV files.

   In this mode, TAB and \\[indent-region] attempt to indent code
based on the position of {} pairs and #-type directives.  The variable
pov-indent-level controls the amount of indentation used inside
arrays and begin/end pairs.  The variable pov-indent-under-declare
determines indent level when you have something like this:
#declare foo =
   some_object {

This mode also provides PoVray keyword fontification using font-lock.
Set pov-fontify-insanely to nil to disable (recommended for large
files!).

\\[pov-complete-word] runs pov-complete-word, which attempts to complete the
current word based on point location.

\\{pov-mode-map}

\\[pov-mode] calls the value of the variable pov-mode-hook 
with no args, if that value is non-nil."
  (interactive)
  (kill-all-local-variables)
  (use-local-map pov-mode-map)
  (pov-setup-syntax-table)
  (make-local-variable 'comment-start)
  (make-local-variable 'comment-start-skip)
  (make-local-variable 'comment-end)
  (make-local-variable 'comment-multi-line)
  (make-local-variable 'comment-column)
  (make-local-variable 'indent-line-function)
  (make-local-variable 'tab-stop-list)
  (make-local-variable 'compile-command)
  (setq compile-command "x-povray")
  (set-syntax-table pov-mode-syntax-table)
  (setq comment-start "// "
	comment-start-skip "/\\*+ *\\|// *"
	comment-end ""
	comment-multi-line nil
	comment-column 60
	indent-line-function 'pov-indent-line
	tab-stop-list pov-tab-stop-list)
  (setq mode-name "PoV")
  (setq major-mode 'pov-mode)
  (pov-setup-font-lock)
  (speedbar-add-supported-extension "\\.pov\\'")  
  (setq imenu-create-index-function 'pov-create-imenu-index)
  (speedbar-add-supported-extension "\\.inc\\'")  
  (setq imenu-create-index-function 'pov-create-imenu-index)

;; Enable support for automatic project switching.
  ;; This feature loads the appropriate project settings whenever
  ;; a user switches from a Java buffer belonging to one project
  ;; to a buffer belonging to another.
  (make-local-hook 'post-command-hook)
  (unless (find 'pov-detect-pov-buffer-activation post-command-hook)
    (add-hook 'post-command-hook 'pov-detect-pov-buffer-activation))

(if (> (pov-count-open-pov-buffers) 1)
    ((if pov-xemacsp
      (pov-insert-menu-in-XEmacs-menubar))))

  ;(if pov-xemacsp
  ;    (pov-insert-menu-in-XEmacs-menubar))

  (make-local-variable 'font-lock-keywords)
  (setq font-lock-keywords pov-font-lock-keywords)
  (run-hooks 'pov-mode-hook)
)

(defun pov-tab ()
  "Command assigned to the TAB key in PoV mode."
  (interactive)
  (if (save-excursion (skip-chars-backward " \t") (bolp))
      (pov-indent-line)
    (save-excursion
      (pov-indent-line))))

(defun pov-indent-line nil
  "Indents a line of PoV code."
  (interactive)
  (beginning-of-line)
  (delete-horizontal-space)
  (if (pov-top-level-p)
      (pov-indent-top-level)
    (if (not (pov-top-level-p))
	(if (pov-in-star-comment-p)
	    (indent-to '2)
	  (if (and (< (point) (point-max))
		   (or
		    (eq ?\) (char-syntax (char-after (point))))
		    (or
		     (looking-at "\\<#\\(end\\|break\\)\\>")
		     (and (looking-at "\\<#else\\>")
			  (not (pov-in-switch-p 0))))))
	      (pov-indent-close)		; indent close-delimiter
	    (pov-indent-in-block))))))	; indent line after open delimiter
  
(defun pov-newline nil
  "Terminate line and indent next line."
  (interactive)
  (newline)
  (pov-indent-line))

(defun pov-in-star-comment-p nil
  "Return true if in a star comment"
  (let ((state
	 (save-excursion
	   (parse-partial-sexp (point-min) (point)))))
    (nth 4 state)))

(defun pov-open nil
  (interactive)
  (insert last-command-char))

(defun pov-close nil
  "Inserts and indents a close delimiter."
  (interactive)
  (insert last-command-char)
  (backward-char 1)
  (pov-indent-close)
  (forward-char 1)
  (blink-matching-open))

(defun pov-indent-close nil
  "Internal function to indent a line containing a close delimiter."
  (if (save-excursion (skip-chars-backward " \t") (bolp))
      (let (x (oldpoint (point)))
	(if (looking-at "#end\\|#else\\|#break")
	    (progn
	      (goto-char (pov-find-begin 0))
	      (if (and (looking-at "#else")
		       (pov-in-switch-p 0))
		  (goto-char (pov-find-begin 0))))
	  (forward-char) (backward-sexp))	;XXX
	(if (and (eq 1 (count-lines (point) oldpoint))
		 (> 1 (- oldpoint (point))))
	    (goto-char oldpoint)
	  (beginning-of-line)
	  (skip-chars-forward " \t")
	  (setq x (current-column))
	  (goto-char oldpoint)
	  (delete-horizontal-space)
	  (indent-to x)))))

(defun pov-indent-in-block nil
  "Indent a line which does not open or close a block."
  (let ((goal (pov-block-start)))
    (setq goal (save-excursion
		 (goto-char goal)
		 (back-to-indentation)
		 (if (bolp)
		     pov-indent-level
		   (back-to-indentation)
		   (+ (current-column) pov-indent-level))))
    (indent-to goal)))

(defun pov-indent-top-level nil
  (if (save-excursion 
	(forward-line -1)
	(looking-at "\\<#declare[ \t]+[0-9a-zA-Z_]+[ \t]*=[ \t]*$"))
	(indent-to pov-indent-under-declare)))

;;; returns nil if at top-level, or char pos of beginning of current block

(defun pov-block-start nil
  "Returns the character position of the character following the nearest
enclosing `{' or `begin' keyword."
  (save-excursion
    (let (open (skip 0))
      (setq open (condition-case nil
		     (save-excursion
		       (backward-up-list 1)
		       (1+ (point)))
		   (error nil)))
      (pov-find-begin open))))

(defun pov-find-begin (start)
  "Search backwards from point to START for enclosing `begin' and returns the
character number of the character following `begin' or START if not found."
  (save-excursion
    (let ((depth 1) match)
      (while (and (> depth 0)
		  (pov-re-search-backward pov-begin-end-re start t))
	(setq depth (if (looking-at pov-end-re)
			(if (and (looking-at "#end")
				 (pov-in-switch-p start))
			    (progn
			      (pov-re-search-backward "\\<#switch\\>" start t)
			      depth)
			  (+ 1 depth))
		      (if (looking-at "\\<#else\\>")
			  (if (pov-in-switch-p start)
			      (1- depth)
			    depth)
			(1- depth)))))
      (if (not (eq 0 depth))
	  start
	(point)))))


(defun pov-in-switch-p (start)
  "Return t if one level under a switch."
  (save-excursion
    (if (looking-at "\\<#end\\>")
	  (pov-re-search-backward pov-begin-end-re start t))
    (beginning-of-line)
    (pov-re-search-backward pov-begin-end-re start t)
    (if (looking-at "\\<#else\\>>") (forward-word -1))
    (while (looking-at "\\<#break\\>")
      (progn
	(pov-re-search-backward "\\<#case\\|#range\\>" start t)
	(pov-re-search-backward pov-begin-end-re start t)))
    (pov-re-search-backward pov-begin-end-re start t)
    (looking-at "\\<#switch\\>")))

(defun pov-top-level-p nil
  "Awful test to see whether we are inside some sort of PoVray block."
  (and (condition-case nil
	   (not (scan-lists (point) -1 1))
	 (error t))
       (not (pov-find-begin nil))))

(defsubst pov-re-search-backward (REGEXP BOUND NOERROR)
  "Like re-search-backward, but skips over matches in comments or strings"
  (set-match-data '(nil nil))
  (while (and
	  (re-search-backward REGEXP BOUND NOERROR)
	  (pov-skip-backward-comment-or-string)
	  (not (set-match-data '(nil nil))))
    ())
  (match-end 0))

(defun pov-autoindent-endblock nil
  "Hack to automatically reindent end, break, and else."
  (interactive)
  (self-insert-command 1)
  (save-excursion
    (forward-word -1)
    (if (looking-at "\\<#else\\|#end\\|#break\\>")
	(pov-indent-line))))

; Taken from verilog-mode.el
(defun pov-skip-backward-comment-or-string ()
 "Return true if in a string or comment"
 (let ((state 
	(save-excursion
	  (parse-partial-sexp (point-min) (point)))))
   (cond
    ((nth 3 state)			;Inside string
     (search-backward "\"")
     t)
    ((nth 7 state)			;Inside // comment
     (search-backward "//")
     t)
    ((nth 4 state)			;Inside /* */ comment
     (search-backward "/*")
     t)
    (t
     nil))))

;;; Completions
(defvar pov-completion-str nil)
(defvar pov-completion-all nil)
(defvar pov-completion-pred nil)
;(defvar pov-completion-buffer-to-use nil)
(defvar pov-completion-flag nil)

(defvar pov-top-level-keywords
  '("global_settings" "camera" "light_source"))

(defvar pov-csg-scope-re 
  "\\<inverse\\|union\\|intersection\\|difference\\|merge\\>")

(defvar pov-solid-primitive-keywords
  '("blob" "box" "cone" "cylinder" "fractal" "height_field" "lathe" "sphere" "superellipsoid" "sor" "text" "torus"))

(defvar pov-blob-keywords
  '("threshold" "cylinder" "sphere" "component" "hierarchy" "sturm"))

(defvar pov-heightfield-keywords
  '("hierarchy" "smooth" "water_level"))


; Julia Fractal

(defvar pov-prism-keywords
  '("linear_sweep" "conic_sweep" "linear_spline" "quadratic_spline" "cubic_spline"))

(defvar pov-patch-primitive-keywords
  '("bicubic_patch" "disc" "smooth_triangle" "triangle" "polygon" "mesh"))

(defvar pov-bicubic-keywords
  '("type" "flatness" "u_steps" "v_steps"))

(defvar pov-infinite-solid-keywords
  '("plane" "cubic" "poly" "quadric" "quartic"))

(defvar pov-csg-keywords
  '("inverse" "union" "intersection" "difference" "merge"))

(defvar pov-light-source-keywords
  '("color" "spotlight" "point_at" "radius" "falloff" "tightness" "area_light" "adaptive" "jitter" "looks_like" "fade_distance" "fade_power" "atmospheric_attenuation"))

(defvar pov-object-modifier-keywords
  '("clipped_by" "bounded_by" "hollow" "no_shadow"))

(defvar pov-transformation-keywords
  '("rotate" "scale" "translate" "matrix"))

(defvar pov-camera-keywords
  '("perspective" "orthographic" "fisheye" "ultra_wide_angle" "omnimax" "panoramic" "cylinder" "location" "look_at" "right" "up" "direction" "sky" "angle" "blur_samples" "aperture" "focal_point" "normal" "rotate" "translate"))

(defvar pov-texture-keywords
  '("pigment" "normal" "finish" "halo" "texture_map" "material_map"))

(defvar pov-pigment-keywords
  '("color" "brick" "checker" "hexagon" "color_map" "gradient" "pigment_map" "pigment" "image_map" "quick_color"))

(defvar pov-normal-keywords
  '("slope_map" "normal_map" "bump_map" "bump_size"))

(defvar pov-finish-keywords
  '("ambient" "diffuse" "brilliance" "phong" "phong_size" "specular" "roughness" "metallic" "reflection" "refraction" "ior" "caustics" "fade_distance" "fade_power" "irid" "crand"))

(defvar pov-halo-keywords
  '("attenuating" "emitting" "glowing" "dust" "constant" "linear" "cubic" "poly" "planar_mapping" "spherical_mapping" "cylindrical_mapping" "box_mapping" "dust_type" "eccentricity" "max_value" "exponent" "samples" "aa_level" "aa_threshold" "jitter" "turbulence" "octaves" "omega" "lambda" "colour_map" "frequency" "phase" "scale" "rotate" "translate"))

(defvar pov-pattern-keywords
  '("agate" "average" "bozo" "granite" "leopard" "marble" "mandel" "onion" "pattern1" "pattern2" "pattern3" "bumpy1" "bumpy2" "bumpy3" "spiral1" "spiral2" "spotted" "wood" "gradient" "crackle" "colour" "checker" "brick" "hexagon" "image_map" "bump_map" "waves" "ripples" "wrinkles" "bumps" "dents" "quilted" ))

(defvar pov-atmosphere-keywords
  '("type" "distance" "scattering" "eccentricity" "samples" "jitter" "aa_threshold" "aa_level" "colour" "color"))

(defvar pov-fog-keywords
 '("fog_type" "distance" "color" "colour" "turbulence" "turb_depth" "omega" "lambda" "octaves" "fog_offset" "fog_alt" "up"))

(defvar pov-global-settings-keywords
  '("adc_bailout" "ambient_light" "assumed_gamma" "hf_gray_16" "irid_wavelength" "max_intersections" "max_trace_level" "number_of_waves" "radiosity"))

(defvar pov-radiosity-keywords
  '("brightness" "count" "distance_maximum" "error_bound" "gray_threshold" "low_error_factor" "minimum_reuse" "nearest_count" "recursion_limit"))

(defvar pov-object-keywords
 '("texture" "pigment" "finish" "halo" "normal"))

(defun pov-string-diff (str1 str2)
  "Return index of first letter where STR1 and STR2 differs."
  (catch 'done
    (let ((diff 0))
      (while t
	(if (or (> (1+ diff) (length str1))
		(> (1+ diff) (length str2)))
	    (throw 'done diff))
	(or (equal (aref str1 diff) (aref str2 diff))
	    (throw 'done diff))
	(setq diff (1+ diff))))))

(defun pov-get-scope nil
  "Return the scope of the POV source at point"
  (interactive)
  (save-excursion
    (if (not (pov-top-level-p))
	(progn
	
    (backward-up-list 1)
    (forward-word -1)
    (cond
     ((looking-at "camera")
      (setq pov-completion-list pov-camera-keywords))
     ((looking-at "texture")
      (setq pov-completion-list (append pov-texture-keywords pov-pattern-keywords)))
     ((looking-at "pigment")
      (setq pov-completion-list (append pov-pigment-keywords pov-pattern-keywords)))
     ((looking-at "normal")
      (setq pov-completion-list (append pov-normal-keywords pov-pattern-keywords)))
     ((looking-at "finish")
      (setq pov-completion-list pov-finish-keywords))
     ((looking-at "halo")
      (setq pov-completion-list pov-halo-keywords))
     ((looking-at "blob")
      (setq pov-completion-list pov-blob-keywords))
     ((looking-at "heightfield")
      (setq pov-completion-list pov-heightfield-keywords))
     ((looking-at "prism")
      (setq pov-completion-list pov-prism-keywords))
     ((looking-at "bicubic")
      (setq pov-completion-list pov-bicubic-keywords))
     ((looking-at "light_source")
      (setq pov-completion-list pov-light-source-keywords))
     ((looking-at "atmosphere")
      (setq pov-completion-list pov-atmosphere-keywords))
     ((looking-at "fog")
      (setq pov-completion-list pov-fog-keywords))
     ((looking-at "global_settings")
      (setq pov-completion-list pov-global-settings-keywords))
     ((looking-at "radiosity")
      (setq pov-completion-list pov-radiosity-keywords))
     ((looking-at pov-csg-scope-re)
      (setq pov-completion-list (append pov-solid-primitive-keywords pov-infinite-solid-keywords pov-object-modifier-keywords pov-csg-keywords)))
     (t
      (setq pov-completion-list (append pov-object-modifier-keywords pov-object-keywords))))
     (setq pov-completion-list (append pov-completion-list pov-transformation-keywords)))
      (setq pov-completion-list (append pov-top-level-keywords pov-solid-primitive-keywords pov-infinite-solid-keywords pov-patch-primitive-keywords pov-csg-keywords)))))

(defun pov-completion (pov-completion-str pov-completion-pred 
					  pov-completion-flag)
  (save-excursion
    (let ((pov-completion-all nil))
      (pov-get-scope)
      (mapcar '(lambda (s)
		 (if (string-match (concat "\\<" pov-completion-str) s)
		     (setq pov-completion-all (cons s pov-completion-all))))
	      pov-completion-list)
      ;; Now we have built a list of all matches. Give response to caller
      (pov-completion-response))))

(defun pov-completion-response ()
  (cond ((or (equal pov-completion-flag 'lambda) (null pov-completion-flag))
	 ;; This was not called by all-completions
	 (if (null pov-completion-all)
	     ;; Return nil if there was no matching label
	     nil
	   ;; Get longest string common in the labels
	   (let* ((elm (cdr pov-completion-all))
		  (match (car pov-completion-all))
		  (min (length match))
		  tmp)
	     (if (string= match pov-completion-str)
		 ;; Return t if first match was an exact match
		 (setq match t)
	       (while (not (null elm))
		 ;; Find longest common string
		 (if (< (setq tmp (pov-string-diff match (car elm))) min)
		     (progn
		       (setq min tmp)
		       (setq match (substring match 0 min))))
		 ;; Terminate with match=t if this is an exact match
		 (if (string= (car elm) pov-completion-str)
		     (progn
		       (setq match t)
		       (setq elm nil))
		   (setq elm (cdr elm)))))
	     ;; If this is a test just for exact match, return nil ot t
	     (if (and (equal pov-completion-flag 'lambda) (not (equal match 't)))
		 nil
	       match))))
	;; If flag is t, this was called by all-completions. Return
	;; list of all possible completions
	(pov-completion-flag
	 pov-completion-all)))

(defun pov-complete-word ()
  "Complete word at current point based on POV syntax."
  (interactive)
  (let* ((b (save-excursion (skip-chars-backward "a-zA-Z0-9_") (point)))
	 (e (save-excursion (skip-chars-forward "a-zA-Z0-9_") (point)))
	 (pov-completion-str (buffer-substring b e))
	 ;; The following variable is used in pov-completion
;	 (pov-buffer-to-use (current-buffer))
	 (allcomp (all-completions pov-completion-str 'pov-completion))
	 (match (try-completion
		 pov-completion-str (mapcar '(lambda (elm)
					(cons elm 0)) allcomp))))

    ;; Delete old string
    (delete-region b e)

    ;; Insert match if found, or the original string if no match
    (if (or (null match) (equal match 't))
	(progn (insert "" pov-completion-str)
	       (message "(No match)"))
      (insert "" match))
      ;; Give message about current status of completion
    (cond ((equal match 't)
	   (if (not (null (cdr allcomp)))
	       (message "(Complete but not unique)")
	     (message "(Sole completion)")))
	  ;; Display buffer if the current completion didn't help 
	  ;; on completing the label.
	  ((and (not (null (cdr allcomp))) (= (length pov-completion-str)
					      (length match)))
	   (with-output-to-temp-buffer "*Completions*"
	     (display-completion-list allcomp))
	     ;; Wait for a keypress. Then delete *Completion*  window
	   (momentary-string-display "" (point))
	   (delete-window (get-buffer-window (get-buffer "*Completions*")))))))


;;; initialize the keymap if it doesn't already exist

(if (null pov-mode-map)
    (progn
      (setq pov-mode-map (make-sparse-keymap))
      (define-key pov-mode-map "{" 'pov-open)
      (define-key pov-mode-map "}" 'pov-close)
      (define-key pov-mode-map "\t" 'pov-tab)
      (define-key pov-mode-map "\r" 'pov-newline)
      (define-key pov-mode-map "\M-\t" 'pov-complete-word)))

;; Hack to reindent end/else/break
(if pov-autoindent-endblocks
    (progn
      (define-key pov-mode-map "e" 'pov-autoindent-endblock)
      (define-key pov-mode-map "k" 'pov-autoindent-endblock)
      (define-key pov-mode-map "d" 'pov-autoindent-endblock)))


;; Define JDE menu for FSF Emacs.
(if (or (not pov-xemacsp) (featurep 'infodock))
    (easy-menu-do-define 'pov-menu 
			 pov-mode-map
			 "Menu for POV."
			 pov-menu)
;
)

(defun pov-insert-menu-in-XEmacs-menubar ()
  "Insert JDE menu in the XEmacs menu bar."
  (if (and 
       (not (featurep 'infodock))
       (not (memq 'infodock c-emacs-features))
       (boundp 'current-menubar)
       current-menubar)
      (if (fboundp 'add-submenu)
	  (add-submenu nil pov-menu)
	(add-menu nil "POV" (cdr pov-menu)))))

(defcustom pov-entering-pov-buffer-hooks '(pov-insert-menu-in-XEmacs-menubar)
"*Lists functions to run when entering a POV source buffer"
  :group 'pov
  :type 'hook)

(setq pov-current-buffer (current-buffer))

(defun pov-detect-pov-buffer-activation ()
"Detects when a user activates a buffer.
If the activated buffer is a Java buffer, runs the 
`pov-entering-pov-buffer' hooks."
  (let ((curr-buff (current-buffer)))
    (if (not
 (equal curr-buff pov-current-buffer))
(progn
  (setq pov-current-buffer curr-buff)
  (if (eq major-mode 'pov-mode)
(run-hooks 'pov-entering-pov-buffer-hooks))))))

(defun pov-count-open-pov-buffers ()
  "Returns non-nil if any POV buffers are open."
  (+
   (count 
    ".pov"
    (buffer-list)
    :test
    (lambda (file-type buffer)
      (let ((file-name (buffer-file-name buffer)))
	(if file-name
	    (string-match file-type file-name)))))
   (count 
    ".inc"
    (buffer-list)
    :test
    (lambda (file-type buffer)
      (let ((file-name (buffer-file-name buffer)))
	(if file-name
	    (string-match file-type file-name)))))
   )
  )


(defun pov-remove-pov-hook ()
  "Removes `pov-detect-pov-buffer-activation-hook' when
all POV source buffers have been closed."
  (unless (> (pov-count-open-pov-buffers) 1)
  (remove-hook 'post-command-hook 'pov-detect-pov-buffer-activation)))

(add-hook 'kill-buffer-hook 'pov-remove-pov-hook)

;; speedbar support. Fully hacked after the Java Development Environment.
(defun pov-make-imenu-patterns ()
  "Makes a replacement for the regular expression indexing  
patterns in imenu, to fit POV's syntax. See `imenu-generic-expression'."
  (let* ((letter "a-zA-Z_$\300-\326\330-\366\370-\377")
	 (digit "0-9")
	 (white-space "\\s-")
	 (optional-white-spaces
	  (concat white-space "*"))
	 (bol "^")
	 (eol "$")
	 (not-comment (concat optional-white-spaces "[^.*/]*"))
	 (anything ".*")

	 (identifier
	  (concat "\\<\\([" letter "][" letter digit "]*\\)\\>"))
	 (identifier-count 1)


	 ;; Pattern for macros
	 (macro1
	  (concat
	   bol
	   not-comment
	   "\\#macro"
	   optional-white-spaces
	   identifier  
	   ))
	 (macro1-count  identifier-count)
	   
	 ;; Pattern for objects
	 (object1
	  (concat
	   bol
	   not-comment
	   "\\#declare"
	   optional-white-spaces
	   identifier
	   ))
	 (object1-count  identifier-count)
	   

	 (exp 
	  (`
	   (
	    ("Objects"
	     (,  object1) (, object1-count))
	    ("Macros"
	     (,  macro1) (, macro1-count))
	    )
	   ))
	 )
    exp))


;;;
;;; Pov index gathering function.
;;;

(defun pov-create-imenu-index ()
;; Based on imenu--generic-function
  "Return an index of the current buffer as an alist.

PATTERN is an alist with elements that look like this: (MENU-TITLE
REGEXP INDEX).

MENU-TITLE is a string used as the title for the submenu or nil if the
entries are not nested.

REGEXP is a regexp that should match a construct in the buffer that is
to be displayed in the menu; i.e., function or variable definitions,
etc.  It contains a substring which is the name to appear in the
menu.  See the info section on Regexps for more information.

INDEX points to the substring in REGEXP that contains the name (of the
function, variable or type) that is to appear in the menu.

For emacs-lisp-mode for example PATTERN would look like:

'((nil \"^\\\\s-*(def\\\\(un\\\\|subst\\\\|macro\\\\|advice\\\\)\\\\s-+\\\\([-A-Za-z0-9]+\\\\)\" 2)
  (\"*Vars*\" \"^\\\\s-*(def\\\\(var\\\\|const\\\\)\\\\s-+\\\\([-A-Za-z0-9]+\\\\)\" 2)
  (\"*Types*\" \"^\\\\s-*(def\\\\(type\\\\|struct\\\\|class\\\\|ine-condition\\\\)\\\\s-+\\\\([-A-Za-z0-9]+\\\\)\" 2))'

Returns an index of the current buffer as an alist.  The elements in
the alist look like: (INDEX-NAME . INDEX-POSITION).  They may also be
nested index lists like (INDEX-NAME . INDEX-ALIST) depending on
pattern.

\(imenu--generic-function PATTERN\)."

  (let* ((patterns (pov-make-imenu-patterns))
	 (case-fold-search nil)
	 (index-alist (list 'dummy))
	 (found nil)
	 (global-regexp 
	  (concat "\\(" 
		  (mapconcat
		   (function (lambda (pattern) (identity (cadr pattern)))) 
		   patterns "\\)\\|\\(") 
		  "\\)"))
	 prev-pos
	 (ppos (point-max)))

    (goto-char (point-max))

    (imenu-progress-message prev-pos 0 t)
    (save-match-data
      (while (re-search-backward global-regexp nil t)
	(imenu-progress-message prev-pos nil t)
        (setq found nil)
	(save-excursion
	  (goto-char (match-beginning 0))
	  (mapcar 
	   (function 
	    (lambda (pat) 
	      (let ((menu-title (car pat))
		    (regexp (cadr pat))
		    (index (caddr pat)))
		    (if (and (not found) ; Only allow one entry;
			     (re-search-forward regexp ppos t))
			(let ((beg (match-beginning index))
			      (end (match-end index)))
			  (setq found t)
			  (setq ppos beg)
			  (push 
			   (cons (buffer-substring-no-properties beg end) beg)
			   (cdr 
			    (or (assoc menu-title index-alist)
				(car (push 
				      (cons menu-title '()) 
				      index-alist))))))))))
	   patterns))))
    (imenu-progress-message prev-pos 100 t)
    (let ((main-element (assq nil index-alist)))
      (nconc (delq main-element (delq 'dummy index-alist)) main-element))))


